@@ -5,19 +5,18 @@
 
 from __future__ import unicode_literals
 
-import base64
 import codecs
-import collections
 import logging
 import os
 import random
-import re
 import socket
 import subprocess
 import sys
+import tempfile
 import time
 
-from . import compat
+from . import control
+from . import core
 
 logger = logging.getLogger(__name__.split('.')[0])
 
@@ -31,14 +30,6 @@ DEFAULT_STARTUP_DELAY = 15
 DEFAULT_SLAPD_DEBUG = 0
 
 
-class LdapError(Exception):
-    """Exceptions for volatildap"""
-
-
-class PathError(LdapError):
-    """Exception for missing paths"""
-
-
 class OpenLdapPaths(object):
     """Collection of Openldap-related paths, distribution dependend."""
 
@@ -57,7 +48,7 @@ class OpenLdapPaths(object):
             fullpath = os.path.join(candidate, needle)
             if os.path.isfile(fullpath):
                 return fullpath
-        raise PathError("Unable to locate file %s; tried %s" % (needle, candidates))
+        raise core.PathError("Unable to locate file %s; tried %s" % (needle, candidates))
 
     _SCHEMA_DIRS = [
         '/etc/ldap/schema',  # Debian
@@ -68,15 +59,13 @@ class OpenLdapPaths(object):
         '/usr/sbin',
         '/usr/bin',
         '/usr/lib/openldap',
+        '/usr/lib64/openldap',
         '/usr/local/sbin',
         '/usr/local/bin',
     ] + os.environ.get('PATH', '').split(':')
 
 
-TLSConfig = collections.namedtuple('TLSConfig', ['root', 'chain', 'certificate', 'key'])
-
-
-class LdapServer(object):
+class LdapServer(core.BaseServer):
     _DATASUBDIR = 'ldif-data'
 
     def __init__(self,
@@ -88,8 +77,10 @@ class LdapServer(object):
                  skip_missing_schemas=False,
                  max_server_startup_delay=DEFAULT_STARTUP_DELAY,
                  port=None,
+                 host='localhost',
                  slapd_debug=DEFAULT_SLAPD_DEBUG,
                  tls_config=None,
+                 control_address=(),
                  ):
 
         self.paths = OpenLdapPaths()
@@ -100,12 +91,20 @@ class LdapServer(object):
         self.initial_data = initial_data or {}
         self.max_server_startup_delay = max_server_startup_delay
         self.port = port or find_available_port()
+        self.host = host
         self.slapd_debug = slapd_debug
         self.tls_config = tls_config
+        self.control = None
 
         self._tempdir = None
         self._process = None
 
+        if control_address:
+            self.control = control.ControlServer(
+                server_address=control_address,
+                ldap_server=self,
+            )
+
     def _generate_password(self):
         chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
         return ''.join(
@@ -131,7 +130,7 @@ class LdapServer(object):
             elif skip_missing_schemas:
                 logger.warning("Unable to locate schema %s at %s", schema, schema_file)
             else:
-                raise PathError("Unable to locate schema %s at %s" % (schema, schema_file))
+                raise core.PathError("Unable to locate schema %s at %s" % (schema, schema_file))
 
     def _configuration_lines(self):
         def quote(base, *args):
@@ -152,14 +151,12 @@ class LdapServer(object):
         yield quote('rootdn %s', self.rootdn)
         yield quote('rootpw %s', self.rootpw)
 
-    def _normalize_dn(self, dn):
-        if not dn.endswith(self.suffix):
-            return '%s,%s' % (dn, self.suffix)
-        else:
-            return dn
-
     def start(self):
+        logger.info("Starting LDAP server")
         try:
+            if self.control is not None:
+                logger.info("Starting control server")
+                self.control.start()
             if self._process is None:
                 self._setup()
                 self._start()
@@ -171,7 +168,14 @@ class LdapServer(object):
             self._shutdown()
             raise
 
+    def wait(self, timeout=None):
+        try:
+            self._process.wait(timeout=timeout)
+        except subprocess.TimeoutExpired as e:
+            raise core.TimeoutExpired(str(e), timeout) from e
+
     def stop(self):
+        logger.info("Shutting down LDAP server")
         self._shutdown()
 
     def add(self, data):
@@ -197,6 +201,9 @@ class LdapServer(object):
         if retcode != 0:
             raise RuntimeError("ldapadd failed with code %d: %s %s" % (retcode, stdout, stderr))
 
+    def add_ldif(self, lines):
+        self.add(core.ldif_to_entries(lines))
+
     def get(self, dn):
         dn = self._normalize_dn(dn)
         logger.info("Fetching data at %s", dn)
@@ -223,30 +230,20 @@ class LdapServer(object):
         if retcode != 0:
             raise RuntimeError("ldapsearch failed with code %d: %r" % (retcode, stderr))
 
-        entries = ldif_to_entries(stdout)
+        entries = core.ldif_to_entries(stdout)
         return entries[dn]
 
+    def get_ldif(self, dn):
+        entry = self.get(dn)
+        lines = self._data_as_ldif({dn: entry})
+        return '\n'.join(lines)
+
     def reset(self):
         """Reset all entries except inital ones."""
+        logger.info("Resetting the LDAP server to its initial data")
         self._clear()
         self._populate()
 
-    def _data_as_ldif(self, data):
-        # Sort by dn length, thus adding parents first.
-        for dn, attributes in sorted(data.items(), key=lambda e: (len(e[0]), e)):
-            yield ldif_encode('dn', self._normalize_dn(dn))
-            for attribute, values in sorted(attributes.items()):
-                for value in values:
-                    yield ldif_encode(attribute, value)
-            yield ''
-
-    @property
-    def uri(self):
-        if self.tls_config:
-            return 'ldaps://localhost.volatildap.org:%d' % self.port
-        else:
-            return 'ldap://localhost:%d' % self.port
-
     @property
     def _datadir(self):
         return os.path.join(self._tempdir.name, self._DATASUBDIR)
@@ -282,7 +279,7 @@ class LdapServer(object):
         }
 
     def _setup(self):
-        self._tempdir = compat.TemporaryDirectory()
+        self._tempdir = tempfile.TemporaryDirectory()
         logger.info("Setting up openldap server in %s", self._tempdir.name)
 
         # Create datadir
@@ -358,7 +355,7 @@ class LdapServer(object):
         if retcode != 0:
             raise RuntimeError("ldapsearch failed with code %d: %r" % (retcode, stderr))
 
-        data = ldif_to_entries(stdout)
+        data = core.ldif_to_entries(stdout)
         dns = data.keys()
         # Remove the furthest first
         dns = sorted(dns, key=lambda dn: (len(dn), dn), reverse=True)
@@ -392,7 +389,7 @@ class LdapServer(object):
 
             s = socket.socket()
             try:
-                s.connect(('localhost', self.port))
+                s.connect((self.host, self.port))
             except socket.error:
                 # Not ready yet, sleep
                 time.sleep(0.5)
@@ -446,58 +443,3 @@ def find_available_port():
     _address, port = s.getsockname()
     s.close()
     return port
-
-
-# Valid characters for a non-base64-encoded LDIF value
-_BASE_LDIF = [
-    chr(i) for i in range(20, 128)
-    if chr(i) not in [' ', '<', ':']
-]
-
-
-def ldif_encode(attr, value):
-    """Encode a attribute: value pair for the LDIF format.
-
-    See RFC2849 for details.
-
-    Rules are:
-    - Text containing only chars <= 127 except control, ' ', '<', ':'
-      is passed as-is
-    - Other text is encoded through UTF-8 and base64-encoded
-    - Binary data is simply base64-encoded
-
-    Returns:
-        A 'key: value' or 'key:: b64value' text line.
-    """
-    if isinstance(value, bytes):
-        return '%s:: %s' % (attr, base64.b64encode(value).decode('ascii'))
-    elif any(c not in _BASE_LDIF for c in value):
-        return '%s:: %s' % (attr, base64.b64encode(value.encode('utf-8')).decode('ascii'))
-    else:
-        return '%s: %s' % (attr, value)
-
-
-def ldif_to_entries(ldif_lines):
-    entries = {}
-    for entry in ldif_lines.decode('ascii').split('\n\n'):
-        if not entry.strip():
-            continue
-
-        attributes = {}
-        for line in entry.split('\n'):
-            if not line.strip():
-                continue
-            m = re.match(r'(\w+)(:?): (.*)', line.strip())
-            if m is None:
-                raise ValueError("Invalid line in ldif output: %r" % line)
-
-            field, is_extended, value = m.groups()
-            if is_extended:
-                value = base64.b64decode(value.encode('ascii'))
-            else:
-                value = value.encode('ascii')
-            attributes.setdefault(field, []).append(value)
-        dns = attributes.get('dn', [b''])
-        assert len(dns) == 1
-        entries[dns[0].decode('utf-8')] = attributes
-    return entries
